﻿using System.Collections.Generic;
using System.Text;
using Microsoft.Xna.Framework;

namespace XELF.Subdivision {

	public class DesignVertex {
		/// <summary>
		/// 位置
		/// </summary>
		public Vector3 Position;
		/// <summary>
		/// 境界用ウェイト値
		/// </summary>
		public float Weight;
		/// <summary>
		/// 演算用の一時的ウェイト値
		/// </summary>
		public float ModulationWeight;
		public int Index;

		public override string ToString() {
			return string.Format("{{Position={0}, Weight={1}, ModulationWeight={2}}}",
				Position, Weight, ModulationWeight);
		}

		public void Lerp(DesignVertex value, float amount) {
			Vector3.Lerp(ref Position, ref value.Position, amount, out Position);
			Weight = MathHelper.Lerp(Weight, value.Weight, amount);
		}
		public void Add(DesignVertex value) {
			Vector3.Add(ref Position, ref value.Position, out Position);
			Weight += value.Weight;
		}
		public void MAdd(DesignVertex value, float scaleFactor) {
			Vector3 position;
			Vector3.Multiply(ref value.Position, scaleFactor, out position);
			Vector3.Add(ref Position, ref position, out Position);
			Weight += value.Weight * scaleFactor;
		}
		public void Add(DesignVertex value, float modificationWeight) {
			Vector3 position;
			Vector3.Multiply(ref value.Position, modificationWeight, out position);
			Vector3.Add(ref Position, ref position, out Position);
			Weight += value.Weight * modificationWeight;
			ModulationWeight += modificationWeight;
		}
		public void Multiply(float scaleFactor) {
			Vector3.Multiply(ref Position, scaleFactor, out Position);
			Weight *= scaleFactor;
		}
		public void Clear() {
			Position = Vector3.Zero;
			Weight = 0;
			ModulationWeight = 0;
		}

		public void Freeze() {
			if (ModulationWeight != 0) Multiply(1 / ModulationWeight);
		}
		public void Modulate() {
			Multiply(ModulationWeight);
		}
	}

	public class FaceVertex {
		public int Vertex;
		public Vector3 Normal;
		public Vector2 TexCoord;

		public override string ToString() {
			return string.Format("{0}", Vertex);
		}

		public void Add(FaceVertex value) {
			Vector3.Add(ref Normal, ref value.Normal, out Normal);
			Vector2.Add(ref TexCoord, ref value.TexCoord, out TexCoord);
		}
		public void Multiply(float scaleFactor) {
			Vector2.Multiply(ref TexCoord, scaleFactor, out TexCoord);
		}
		/// <summary>
		/// 法線を正規化し、テクスチャ座標をスケールする。
		/// </summary>
		/// <param name="scaleFactor">テクスチャ座標に乗算する値。</param>
		public void Normalize(float scaleFactor) {
			Normal.Normalize();
			Vector2.Multiply(ref TexCoord, scaleFactor, out TexCoord);
		}
		public void Clear() {
			Normal = Vector3.Zero;
			TexCoord = Vector2.Zero;
		}
	}

	public class Face {
		public List<FaceVertex> Vertices;
		public List<Edge> Edges;

		public override string ToString() {
			StringBuilder s = new StringBuilder();
			if (Vertices.Count > 0) {
				s.Append(Vertices[0].Vertex);
			}
			for (int i = 1; i < Vertices.Count; i++) {
				s.Append('-');
				s.Append(Vertices[i].Vertex);
			}
			return s.ToString();
		}
	}

	public struct EdgeKey {
		public readonly int From;
		public readonly int To;

		public EdgeKey(int a, int b) {
			if (a < b) {
				From = a;
				To = b;
			} else {
				From = b;
				To = a;
			}
		}
		public override string ToString() {
			return string.Format("{0}-{1}", From, To);
		}
	}

	public class Edge {
		public DesignVertex From, To;
		public float Weight;
		public int Index;

		public readonly List<Face> Faces = new List<Face>();

		public void Midpoint(ref DesignVertex result) {
			result.Clear();
			result.Add(From);
			result.Add(To);
			result.Multiply(0.5f);
		}

		public override string ToString() {
			return string.Format("{0}-{1}", From, To);
		}
	}

	public class DesignMesh {
		public List<DesignVertex> Vertices;
		public List<Face> Faces;
		public List<Edge> Edges = new List<Edge>();

		readonly Dictionary<EdgeKey, Edge> edgeDictionary = new Dictionary<EdgeKey, Edge>();

		/// <summary>
		/// 頂点自身のインデックスを振り直す。
		/// </summary>
		public void RenumberVertices() {
			for (int i = 0; i < Vertices.Count; i++) {
				Vertices[i].Index = i;
			}
		}
		/// <summary>
		/// 辺の情報を頂点と面から作る。
		/// </summary>
		public void BuildEdges() {
			Edges.Clear();

			foreach (var face in Faces) {
				var vn = face.Vertices.Count;
				var previousVertex = face.Vertices[vn - 1];
				foreach (var vertex in face.Vertices) {
					var key = new EdgeKey(previousVertex.Vertex, vertex.Vertex);

					Edge edge;
					if (!edgeDictionary.TryGetValue(key, out edge)) {
						edge = new Edge { From = Vertices[previousVertex.Vertex], To = Vertices[vertex.Vertex], };
						edgeDictionary.Add(key, edge);
						Edges.Add(edge);
					}
					edge.Faces.Add(face);
					previousVertex = vertex;
				}
			}

			for (int i = 0; i < Edges.Count; i++) {
				Edges[i].Index = i;
			}
		}

		public Edge Edge(int vertex0, int vertex1) {
			return edgeDictionary[new EdgeKey(vertex0, vertex1)];
		}

		public DesignMesh(List<DesignVertex> vertices, List<Face> faces) {
			Vertices = vertices;
			Faces = faces;
			RenumberVertices();
			BuildEdges();
		}
	}

}
