﻿using System;
using System.Collections.Generic;

namespace XELF.Subdivision {

	public static class SubdivisionHelper {
		public static void FacePoint(this Face face, List<DesignVertex> vertices, DesignVertex result) {
			result.Clear();
			foreach (var vertex in face.Vertices) {
				result.Add(vertices[vertex.Vertex]);
			}
			var scale = 1.0f / face.Vertices.Count;
			result.Multiply(scale);
		}
		public static void FacePoint(this Face face, List<DesignVertex> vertices, ref FaceVertex result) {
			result.Clear();
			vertices[result.Vertex].Clear();
			foreach (var vertex in face.Vertices) {
				result.Add(vertex);
				vertices[result.Vertex].Add(vertices[vertex.Vertex]);
			}
			var scale = 1.0f / face.Vertices.Count;
			result.Multiply(scale);
			vertices[result.Vertex].Multiply(scale);
		}
		public static void EdgePoint(this Edge edge, List<DesignVertex> vertices, ref DesignVertex result) {
			result.Clear();
			result.Add(edge.From);
			result.Add(edge.To);
			result.Multiply(0.5f);
		}
		/// <summary>
		/// 値をウェイトで除算して、演算用ウェイトはそのままにする。
		/// </summary>
		/// <param name="vertices"></param>
		/// <param name="start"></param>
		/// <param name="count"></param>
		public static void Freeze(this List<DesignVertex> vertices, int start, int count) {
			for (int i = start, e = start + count; i < e; i++) {
				vertices[i].Freeze();
			}
		}
		/// <summary>
		/// 値をウェイトで乗算して、演算用ウェイトはそのままにする。
		/// </summary>
		/// <param name="vertices"></param>
		/// <param name="start"></param>
		/// <param name="count"></param>
		public static void Modulate(this List<DesignVertex> vertices, int start, int count) {
			for (int i = start, e = start + count; i < e; i++) {
				vertices[i].Modulate();
			}
		}
		/// <summary>
		/// 初期値に戻す。
		/// </summary>
		/// <param name="vertices"></param>
		/// <param name="start"></param>
		/// <param name="count"></param>
		public static void Clear(this List<DesignVertex> vertices, int start, int count) {
			for (int i = start, e = start + count; i < e; i++) {
				vertices[i].Clear();
			}
		}
		public static void Add(this List<DesignVertex> destinationVertices, int desitinationStart,
			List<DesignVertex> sourceVertices, int sourceStart, int count) {
			for (int i = desitinationStart, e = desitinationStart + count, j = sourceStart; i < e; i++, j++) {
				destinationVertices[i].Add(sourceVertices[j]);
			}
		}
		public static void LerpWithSourceWeight(this List<DesignVertex> destinationVertices, int desitinationStart,
					List<DesignVertex> sourceVertices, int sourceStart, int count) {
			for (int i = desitinationStart, e = desitinationStart + count, j = sourceStart; i < e; i++, j++) {
				destinationVertices[i].Lerp(sourceVertices[j], sourceVertices[j].Weight);
			}
		}

		public static void Mid(this FaceVertex result, FaceVertex a, FaceVertex b) {
			result.Clear();
			result.Add(a);
			result.Add(b);
			result.Normalize(0.5f);
		}
		public static void CopyUniqueProperties(this FaceVertex result, FaceVertex source) {
			result.Normal = source.Normal;
			result.TexCoord = source.TexCoord;
		}
		/// <summary>
		/// 面頂点固有情報の平均を得る。ただし、法線は正規化されたものを得る。
		/// </summary>
		/// <param name="face"></param>
		/// <param name="result"></param>
		public static void AverageVerticesUniqueProperies(this Face face, ref FaceVertex result) {
			foreach (var vertex in face.Vertices) {
				result.Add(vertex);
			}
			var scale = 1.0f / face.Vertices.Count;
			result.Normalize(scale);
		}
	}

	public class SubdivisionService {
		readonly List<DesignMesh> meshes = new List<DesignMesh>();

		public SubdivisionService(DesignMesh source) {
			this.meshes.Add(source);
		}
		public DesignMesh this[int level] {
			get {
				if (level == 0) {
					return meshes[0];
				}
				for (int i = 0; i < level; i++) {
					if (i + 1 >= meshes.Count) Subdivide(i);
				}
				return meshes[level];
			}
		}

		void Subdivide(int level) {
			var source = meshes[level];

			if (level + 1 >= meshes.Count)
				meshes.Add(new DesignMesh(new List<DesignVertex>(), new List<Face>()));

			var target = meshes[level + 1];

			// 境界の調整
			foreach (var edge in source.Edges) {
				if (edge.Faces.Count <= 1) {
					edge.From.Weight = 1;
					edge.To.Weight = 1;
					edge.Weight = 1;
				}
			}

			int vertex0Count; // V0の数
			int vertexFacePointStart; // verticesの面の点開始インデックス
			int vertexEdgePointStart; // verticesの辺の点開始インデックス
			int vertexVertexPointStart; // verticesの頂点の点開始インデックス
			int vertexTemporalStart; // verticesの一時データ開始インデックス

			vertex0Count = source.Vertices.Count;

			vertexVertexPointStart = 0;
			vertexEdgePointStart = vertexVertexPointStart + source.Vertices.Count;
			vertexFacePointStart = vertexEdgePointStart + source.Edges.Count;
			vertexTemporalStart = vertexFacePointStart + source.Faces.Count;

			// 細分化した頂点（共有成分）を作る（作る数 = 面の数 + 辺の数 + 頂点の数 + (一時利用頂点の数)）
			var numberFaceVertex1s = source.Faces.Count + source.Edges.Count + source.Vertices.Count * 2;
			for (int faceIndex = 0; faceIndex < numberFaceVertex1s; faceIndex++) {
				target.Vertices.Add(new DesignVertex());
			}

			// face pointsの位置
			for (int faceIndex = 0; faceIndex < source.Faces.Count; faceIndex++) {
				source.Faces[faceIndex].FacePoint(source.Vertices, target.Vertices[vertexFacePointStart + faceIndex]);
			}
			// edge pointsの位置 ( e1 = (隣接f1の平均 + 隣接v0の平均) / 2 )
			for (int f = 0; f < source.Faces.Count; f++) { // 隣接f1の加算
				var face = source.Faces[f];

				var previousVertex = face.Vertices[face.Vertices.Count - 1].Vertex;
				for (int v = 0; v < face.Vertices.Count; v++) {
					var vertex = face.Vertices[v].Vertex;
					var edge = source.Edge(previousVertex, vertex);
					int e = edge.Index;
					var edgePoint = target.Vertices[vertexEdgePointStart + e];
					edgePoint.Add(target.Vertices[vertexFacePointStart + f], 1 - edge.Weight); // e1 += f1
					previousVertex = vertex;
				}
			}

			for (int e = 0; e < source.Edges.Count; e++) { // 隣接v0の加算
				var edge = source.Edges[e];
				target.Vertices[vertexEdgePointStart + e].Add(edge.From, 1);
				target.Vertices[vertexEdgePointStart + e].Add(edge.To, 1);
			}

			target.Vertices.Freeze(vertexEdgePointStart, source.Edges.Count);

			// vertex pointsの位置
			// = 隣接f1の平均/n + 2*隣接e1の平均/n + 元のv0*(n-3)/n
			// = (隣接f1の平均 + 隣接v0の平均 + 元のv0*(n-2)) / n
			{
				// v0周りの隣接v0の平均を加算
				for (int e = 0; e < source.Edges.Count; e++) {
					Edge edge = source.Edges[e];
					target.Vertices[vertexVertexPointStart + edge.From.Index].Add(source.Vertices[edge.To.Index], 1);
					target.Vertices[vertexVertexPointStart + edge.To.Index].Add(source.Vertices[edge.From.Index], 1);
				}

				target.Vertices.Freeze(vertexVertexPointStart, vertex0Count); // v1 = 隣接v0の平均

				// v0 * (n - 2)
				for (int v = 0; v < vertex0Count; v++) {
					var w = target.Vertices[vertexVertexPointStart + v].ModulationWeight;
					w = Math.Max(1, w - 2);
					target.Vertices[vertexTemporalStart + v].MAdd(source.Vertices[v], w);
				}

				target.Vertices.Add(vertexVertexPointStart, target.Vertices, vertexTemporalStart, vertex0Count); // v1 += t1
				target.Vertices.Clear(vertexTemporalStart, vertex0Count);

				// v0周りのf1の平均を加算
				for (int f = 0; f < source.Faces.Count; f++) {
					var face = source.Faces[f];
					for (int v = 0; v < face.Vertices.Count; v++) {
						target.Vertices[vertexTemporalStart + face.Vertices[v].Vertex].Add(
							target.Vertices[vertexFacePointStart + f], 1);
					}
				}

				target.Vertices.Freeze(vertexTemporalStart, vertex0Count); // t1 = f1の平均

				target.Vertices.Add(vertexVertexPointStart, target.Vertices, vertexTemporalStart, vertex0Count);

				target.Vertices.Freeze(vertexVertexPointStart, vertex0Count); // v1 *= (1 / n)
				// 頂点ウェイトの効果
				target.Vertices.LerpWithSourceWeight(vertexVertexPointStart, source.Vertices, 0, vertex0Count);
			}

			// 面を作る
			FaceVertex facePointSettings = new FaceVertex();
			FaceVertex previousEdgePointSettings = new FaceVertex();
			FaceVertex edgePointSettings = new FaceVertex();
			for (int f = 0; f < source.Faces.Count; f++) {
				var face0 = source.Faces[f];

				face0.AverageVerticesUniqueProperies(ref facePointSettings);

				var v0 = face0.Vertices[face0.Vertices.Count - 2];
				var v1 = face0.Vertices[face0.Vertices.Count - 1];
				var previousVertex = v0.Vertex;
				var vertex = v1.Vertex;
				previousEdgePointSettings.Mid(v0, v1);
				var previousEdgePoint = vertexEdgePointStart + source.Edge(previousVertex, vertex).Index;
				previousVertex = vertex;
				var faceIndex = vertexFacePointStart + f;
				v0 = v1;

				for (int e = 0; e < face0.Vertices.Count; e++) {
					v1 = face0.Vertices[e];
					vertex = v1.Vertex;
					var edgePoint = vertexEdgePointStart + source.Edge(previousVertex, vertex).Index;

					edgePointSettings.Mid(v0, v1);

					var face1 = new Face
					{
						Vertices = new List<FaceVertex> {
							new FaceVertex { Vertex = previousEdgePoint, }, // e1
							new FaceVertex { Vertex = previousVertex, }, // v1
							new FaceVertex { Vertex = edgePoint, }, // e1'
							new FaceVertex { Vertex = faceIndex, }, // f1
						}
					};
					target.Faces.Add(face1);

					face1.Vertices[0].CopyUniqueProperties(previousEdgePointSettings);
					face1.Vertices[1].CopyUniqueProperties(v0);
					face1.Vertices[2].CopyUniqueProperties(edgePointSettings);
					face1.Vertices[3].CopyUniqueProperties(facePointSettings);

					previousVertex = vertex;
					previousEdgePoint = edgePoint;
					Swap(ref previousEdgePointSettings, ref edgePointSettings);
					v0 = v1;
				}
			}

			target.RenumberVertices();
			target.BuildEdges();

			target.Vertices.RemoveRange(vertexTemporalStart, vertex0Count);

			foreach (var vertex in target.Vertices) {
				vertex.ModulationWeight = 0;
				vertex.Weight = 0;
			}
		}

		public static void Swap(ref FaceVertex a, ref FaceVertex b) {
			FaceVertex c = b;
			b = a;
			a = c;
		}
	}

}
